#!/usr/bin/python

# wave_futz.py
#       January 12, 2009        bar
#
#       Run some tests on a 16 bit mono .wav file.
#
#       The input file is named, test_16_mono.wav.
#       We output a bunch of x_*.wav files.
#       We read the first output files back in and transform them to other output files.
#
#       The input file can be 8 or 16 bit, mone or stereo.
#       But it's 16-mono to test reading from 16-mono.
#       If it's stereo, we only use the left channel.
#       If it's 8 bit, sensitive ears won't distinguish stuff that's been converted to 8-bit and back out to 16.
#
#       I've looked at 8 and 16 bit files, mono and stereo.
#       I have not looked at 32 bit files.
#       I have not messed with the _convert() logic in wave.py, whatever it is. (I presume it's for ADPCM or one of the xLaws' companding.)
#       I have not thought about incremental read and writes of samples, nor intermixing sample reading/writing with direct frame ops.
#           The latter is just weird, anyway.
#
#


import  array
import  wave
import  struct


def readsamples(me, nframes) :
    """
        Return an array of number-of-channels normalized int sample arrays read from the file.

        To include in wave.py, change 'me' to 'self' and (is think only) squeeze the colons to the left.
    """

    wav = me.readframes(nframes)

    if    me.getsampwidth() == 4 :
        wav     = struct.unpack("<%ul" % (len(wav) / 4), wav)       # !!!! untested
    elif  me.getsampwidth() == 2 :
        wav     = struct.unpack("<%uh" % (len(wav) / 2), wav)
    else :
        wav     = struct.unpack("%uB"  %  len(wav),      wav)
        wav     = [ s - 128 for s in wav ]

    nc          = me.getnchannels()
    if  nc > 1  :
        wavs    = []
        for c in xrange(nc) :
            wavs.append([ wav[si] for si in xrange(c, len(wav), nc) ])
        pass
    else :
        wavs    = [ wav ]

    return(wavs)



def writesamples(me, wavs) :
    """
        Write samples to the wave file.
        'wavs' looks like this:
               [ left_channel_samples,  right_channel_samples ]
            or [ left_channel_samples                         ]
            or   mono_samples
        This routine calls setnchannels() from information about 'wavs' length.

        Note: I have no idea what might happen if multiple calls are done to "stream" to the .wav file.

        To include in wave.py, change 'me' to 'self' and (is think only) squeeze the colons to the left.
    """

    if  wavs :

        if  len(wavs) not in [ 1, 2, 4 ] :
            wavs        = [ wavs ]

        me.setnchannels(len(wavs))

        if  len(wavs)   > 1 :
            wav         = []
            for w in zip(*wavs):
                wav    += w
            pass
        else :
            wav         = wavs[0]

        if    me.getsampwidth() == 4 :
            ws  = array.array('l', [ s       for s in wav ])        # !!!! untested
        elif  me.getsampwidth() == 2 :
            ws  = array.array('h', [ s       for s in wav ])
        else :
            ws  = array.array('B', [ s + 128 for s in wav ])

        ws  = ws.tostring()

        me.writeframes(ws)                                          # we want all the samples in writeframes() format so that _convert can be called on them

    pass






def normalize_samples(samps, sample_width) :
    try :
        samps[0] + 1
    except TypeError :
        samps   = [ ord(s) for s in samps ]                         # in case samps is a string of, presumably, bytes, as it may be in the program this code comes from

    avg         = sum([ s for s in samps ]) / len(samps)
    samps       = [ s - avg for s in samps ]                        # take out the DC by subtracting the simple average from all the samples (this code was used for an ADC board)

    mx_amp      = 127.0
    if  sample_width == 2 :
        mx_amp  = 32767.0
    mx  = max(abs(min(samps)), max(samps))
    if  mx > 0  :
        mxm     = mx_amp / mx
        samps   = [ int(s * mxm) for s in samps ]                   # scale the samples to full amplitude

    return(samps)


def write_wave(ofile_name, wavs, sample_rate, sample_width = 2) :
    if  wavs :
        if  len(wavs) > 4 :
            wavs    =   normalize_samples(wavs, sample_width)
        else :
            wavs    = [ normalize_samples(wav,  sample_width) for  wav in wavs ]

        fo          = wave.open(ofile_name, "wb")
        fo.setsampwidth(sample_width)
        fo.setframerate(sample_rate)
        writesamples(fo, wavs)
        fo.close()
        del(fo)
    pass



def read_wave(ifile_name) :
    try :
        fi      = wave.open(ifile_name, "rb")
        sr      = fi.getframerate()
        wavs    = readsamples(fi, fi.getnframes())
        fi.close()
        del(fi)

        return( [ sr, wavs ] )
    except IOError :
        pass

    return( [ 0, None ] )




if __name__ == '__main__' :

    ( sr, wavs )    = read_wave("test_16_mono.wav")                                         # actually, this file can be a stereo file, 8 or 16 bit, but we ignore the right channel
    if  wavs        :
        wv          = [ s for s in wavs[0] ]
        wv.reverse()


        write_wave("x_16_stereo_right_channel_backward.wav", [ wavs[0], wv      ], sr, 2)
        write_wave("x_8_mono.wav",                             wavs[0],            sr, 1)
        write_wave("x_8_stereo_left_channel_backward.wav",   [ wv,      wavs[0] ], sr, 1)


        ( sr, wavs )    = read_wave("x_16_stereo_right_channel_backward.wav")
        wav             = [ wavs[0][i] + wavs[1][i] for i in xrange(len(wavs[0])) ]         # combine the two tracks to one mono
        write_wave("x_8_mono_forward_backward_together.wav", [ wav              ], sr, 1)   # note: write_wave() will scale the samples so it's ok that they are out of range


        ( sr, wavs )    = read_wave("x_8_mono.wav")
        fs              = 0.0
        wav             = []
        for s in wavs[0] :
            fs          = ((fs * 127) + s) / 128;
            wav.append(fs)                                                                  # cheeseball iir filter to mess with the sound
        write_wave("x_16_mono_filtered_from_8_bit.wav",      [ wav              ], sr, 2)   # note: write_wave() will scale the samples so they won't be really quiet. And it will convert them to ints


        ( sr, wavs )    = read_wave("x_8_stereo_left_channel_backward.wav")
        wv              = [ s for s in wavs[0] ]
        wv.reverse()
        fs              = 0.0
        wav             = []
        for s in wavs[1] :
            fs          = ((fs * 127) + s) / 128;
            wav.append(fs)                                                                  # cheeseball iir filter to mess with the sound
        write_wave("x_16_stereo_right_channel_filtered.wav", [ wv,      wav     ], sr, 2)

    pass


#
#
# eof
